@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2007-2020 All Rights Reserved.
+%% Copyright Ericsson AB 2007-2022 All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -19,10 +19,46 @@
 %%
 
 %%----------------------------------------------------------------------
-%% Purpose: Help funtions for handling certificat verification.
+%% Purpose: Help functions for handling certificate verification.
 %% The path validation defined in ssl_handshake.erl that mainly
 %% calls functions in this module is described in RFC 3280. 
-%%----------------------------------------------------------------------
+%% The basic verification checks are done by
+%% public_key:pkix_path_validation/3
+%%
+%% TLS code handles construction of alternative certificate paths
+%% that can be used as input to public_key:pkix_path_validation/3
+%% to try and find one path that is considered valid.
+%%
+%% The TLS protocol will send certificate chains that should consist
+%% of [PeerCert, CA0 ... CAN, ROOTCert]. The path will be the reverse of the
+%% chain and ROOTCert is the trusted anchor in the path validation.
+%% Also to complicate matters ROOTCerts can be left out of the sent chain.
+%% However due to configuration error and workarounds for certificate
+%% renewal purposes we have to handle chains that may be:
+%%
+%% * Unordered - PeerCert will still be first but the other certs
+%% may be arbitrarily order. Ex: [PeerCert, CAN ...  ROOTCert, CA0]
+%%
+%% * Partial - User decides to put the trust in an intermediate CA
+%% that is this intermediate CA must be part of the original chain
+%% and is used as the trusted anchor cert instead of the ROOT certificate.
+%% Ex: [PeerCert, CA0 ...CAN-1] instead of  [PeerCert, CA0 ...CAN, ROOTCert]
+%%
+%% * Incomplete - The chain sent is missing one or more certificates
+%% but if we have the missing certificates in our trust store we can recreate
+%% the chain. Ex:  [PeerCert, CA0 ...CAN-1] and CN and ROOTCert is in our
+%% trust store.
+%%
+%% * Extraneous - Contain extra certificates that are so called cross signed
+%% to enable construction of different cert chains depending on what is in the
+%% trust store. Used to phase out certificates that are expiring. So that
+%% a window can be created when there is an old about to expire cert and
+%% a new replacing cert that can coexist.
+%% Ex:  [PeerCert, CA0 ...CAN, CAN', ROOTCert]
+%%
+%% * Cross signed ROOT - Means looking for alternative paths using possible
+%% alternative ROOT certs if the original ROOT cert has expired or is unknown. 
+%% ----------------------------------------------------------------------
 
 -module(ssl_certificate).
 
@@ -32,26 +68,30 @@
 -include_lib("public_key/include/public_key.hrl"). 
 
 -export([trusted_cert_and_paths/4,
-	 certificate_chain/3,
-	 certificate_chain/4,
-	 file_to_certificats/2,
-	 file_to_crls/2,
-	 validate/3,
-	 is_valid_extkey_usage/2,
-	 is_valid_key_usage/2,
-	 select_extension/2,
-	 extensions_list/1,
-	 public_key_type/1,
-	 foldl_db/3
+         certificate_chain/3,
+         certificate_chain/5,
+         file_to_certificats/2,
+         file_to_crls/2,
+         validate/3,
+         is_valid_extkey_usage/2,
+         is_valid_key_usage/2,
+         select_extension/2,
+         extensions_list/1,
+         public_key_type/1,
+         foldl_db/3,
+         find_cross_sign_root_paths/4,
+         handle_cert_auths/4,
+         available_cert_key_pairs/1,
+         available_cert_key_pairs/2
 	]).
- 
+
 %%====================================================================
 %% Internal application API
 %%====================================================================
 
 %%--------------------------------------------------------------------
 -spec trusted_cert_and_paths([der_cert()], db_handle(), certdb_ref(), fun()) ->
-				   [{der_cert() | unknown_ca | invalid_issuer | selfsigned_peer, [der_cert()]}].
+          [{public_key:combined_cert() | unknown_ca | invalid_issuer | selfsigned_peer, [public_key:combined_cert()]}].
 %%
 %% Description: Construct input to public_key:pkix_path_validation/3,
 %% If the ROOT cert is not found {bad_cert, unknown_ca} will be returned
@@ -61,20 +101,22 @@
 %% Note: Path = lists:reverse(Chain) -- Root, that is on the peer cert 
 %% always comes first in the chain but last in the path.
 %%--------------------------------------------------------------------
-trusted_cert_and_paths([Peer] = Chain,  CertDbHandle, CertDbRef, PartialChainHandler) ->
+trusted_cert_and_paths([Peer],  CertDbHandle, CertDbRef, PartialChainHandler) ->
     OtpCert = public_key:pkix_decode_cert(Peer, otp),
+    Chain = [#cert{der=Peer, otp=OtpCert}],
     case public_key:pkix_is_self_signed(OtpCert) of
         true ->
-            [{selfsigned_peer, [Peer]}];
+            [{selfsigned_peer, Chain}];
         false ->
-            [handle_incomplete_chain(Chain, PartialChainHandler, {unknown_ca, [Peer]},
+            [handle_incomplete_chain(Chain, PartialChainHandler, {unknown_ca, Chain},
                                      CertDbHandle, CertDbRef)]
     end;
-trusted_cert_and_paths(Chain,  CertDbHandle, CertDbRef, PartialChainHandler) ->
+trusted_cert_and_paths(Chain0,  CertDbHandle, CertDbRef, PartialChainHandler) ->
     %% Construct possible certificate paths from the chain certificates.
     %% If the chain contains extraneous certificates there could be
-    %% more than one possible path such chains might be used to phase out 
-    %% an old certificate. 
+    %% more than one possible path such chains might be used to phase out
+    %% an old certificate.
+    Chain = [#cert{der=Der,otp=public_key:pkix_decode_cert(Der, otp)} || Der <- Chain0],
     Paths = paths(Chain, CertDbHandle),
     lists:map(fun(Path) ->
                       case handle_partial_chain(Path, PartialChainHandler, CertDbHandle, CertDbRef) of
@@ -83,37 +125,59 @@ trusted_cert_and_paths(Chain,  CertDbHandle, CertDbRef, PartialChainHandler) ->
                                                       PartialChainHandler, 
                                                       Result,
                                                       CertDbHandle, CertDbRef);
-                          Result ->
+                          {_Root, _NewChain} = Result ->
                               Result
                       end
               end, Paths).
 %%--------------------------------------------------------------------
--spec certificate_chain(undefined | binary() | #'OTPCertificate'{} , db_handle(), certdb_ref() | {extracted, list()}) ->
-			  {error, no_cert} | {ok, der_cert() | undefined, [der_cert()]}.
+-spec certificate_chain([] | binary() | #'OTPCertificate'{} , db_handle(),
+                        certdb_ref() | {extracted, list()}) ->
+          {error, no_cert} | {ok, der_cert() | undefined, [der_cert()]}.
 %%
 %% Description: Return the certificate chain to send to peer.
 %%--------------------------------------------------------------------
-certificate_chain(undefined, _, _) ->
+certificate_chain([], _, _) ->
     {error, no_cert};
-certificate_chain(OwnCert, CertDbHandle, CertsDbRef) when is_binary(OwnCert) ->
-    ErlCert = public_key:pkix_decode_cert(OwnCert, otp),
-    certificate_chain(ErlCert, OwnCert, CertDbHandle, CertsDbRef, [OwnCert], []);
-certificate_chain(OwnCert, CertDbHandle, CertsDbRef) ->
-    DerCert = public_key:pkix_encode('OTPCertificate', OwnCert, otp),
-    certificate_chain(OwnCert, DerCert, CertDbHandle, CertsDbRef, [DerCert], []).
-
+certificate_chain(DerCert, CertDbHandle, CertsDbRef) when is_binary(DerCert) ->
+    ErlCert = public_key:pkix_decode_cert(DerCert, otp),
+    Cert = #cert{der=DerCert, otp=ErlCert},
+    {ok, Root, Chain} = build_certificate_chain(Cert, CertDbHandle, CertsDbRef, [Cert], []),
+    chain_result(Root, Chain, encoded);
+certificate_chain(#'OTPCertificate'{} = OtpCert, CertDbHandle, CertsDbRef) ->
+    DerCert = public_key:pkix_encode('OTPCertificate', OtpCert, otp),
+    Cert = #cert{der=DerCert, otp=OtpCert},
+    {ok, Root, Chain} = build_certificate_chain(Cert, CertDbHandle, CertsDbRef, [Cert], []),
+    chain_result(Root, Chain, encoded);
+certificate_chain(#cert{} = Cert, CertDbHandle, CertsDbRef) -> 
+    {ok, Root, Chain} = build_certificate_chain(Cert, CertDbHandle, CertsDbRef, [Cert], []),
+    chain_result(Root, Chain, encoded).
 %%--------------------------------------------------------------------
--spec certificate_chain(undefined | binary() | #'OTPCertificate'{} , db_handle(), certdb_ref() | {extracted, list()}, [der_cert()]) ->
-			  {error, no_cert} | {ok, der_cert() | undefined, [der_cert()]}.
+-spec certificate_chain(binary() | #'OTPCertificate'{} , db_handle(), certdb_ref() | 
+                        {extracted, list()}, [der_cert()], encoded | decoded | both) ->
+          {ok,
+           der_cert() | #'OTPCertificate'{} | undefined,
+           [der_cert() |  #'OTPCertificate'{}]} |
+          {ok,
+           {der_cert() | undefined,  [der_cert()]},
+           {#'OTPCertificate'{} | undefined, [#'OTPCertificate'{}]}
+          }.
 %%
-%% Description: Create certificate chain with certs from 
+%% Description: Create certificate chain with certs from Candidates
 %%--------------------------------------------------------------------
-certificate_chain(Cert, CertDbHandle, CertsDbRef, Candidates) when is_binary(Cert) ->
-    ErlCert = public_key:pkix_decode_cert(Cert, otp),
-    certificate_chain(ErlCert, Cert, CertDbHandle, CertsDbRef, [Cert], Candidates);
-certificate_chain(Cert, CertDbHandle, CertsDbRef, Candidates) ->
-    DerCert = public_key:pkix_encode('OTPCertificate', Cert, otp),
-    certificate_chain(Cert, DerCert, CertDbHandle, CertsDbRef, [DerCert], Candidates).
+certificate_chain(DerCert, CertDbHandle, CertsDbRef, Candidates, Type) when is_binary(DerCert) ->
+    ErlCert = public_key:pkix_decode_cert(DerCert, otp),
+    Cert = #cert{der=DerCert, otp=ErlCert},
+    {ok, Root, Chain} = build_certificate_chain(Cert, CertDbHandle, CertsDbRef, [Cert], Candidates),
+    chain_result(Root, Chain, Type);
+certificate_chain(#'OTPCertificate'{} = OtpCert, CertDbHandle, CertsDbRef, Candidates, Type) ->
+    DerCert = public_key:pkix_encode('OTPCertificate', OtpCert, otp),
+    Cert = #cert{der=DerCert, otp=OtpCert},
+    {ok, Root, Chain} = build_certificate_chain(Cert, CertDbHandle, CertsDbRef, [Cert], Candidates),
+    chain_result(Root, Chain, Type);
+certificate_chain(#cert{} = Cert, CertDbHandle, CertsDbRef, Candidates, Type) -> 
+    {ok, Root, Chain} = build_certificate_chain(Cert, CertDbHandle, CertsDbRef, [Cert], Candidates),
+    chain_result(Root, Chain, Type).
+                
 %%--------------------------------------------------------------------
 -spec file_to_certificats(binary(), term()) -> [der_cert()].
 %%
@@ -134,9 +198,7 @@ file_to_crls(File, DbHandle) ->
 
 %%--------------------------------------------------------------------
 -spec validate(term(), {extension, #'Extension'{}} | {bad_cert, atom()} | valid | valid_peer,
-	       term()) -> {valid, term()} |
-			  {fail, tuple()} |
-			  {unknown, term()}.
+	       term()) -> {valid, term()} | {fail, tuple()} | {unknown, term()}.
 %%
 %% Description:  Validates ssl/tls specific extensions
 %%--------------------------------------------------------------------
@@ -150,6 +212,8 @@ validate(_,{extension, #'Extension'{extnID = ?'id-ce-extKeyUsage',
     end;
 validate(_, {extension, _}, UserState) ->
     {unknown, UserState};
+validate(Issuer, {bad_cert, cert_expired}, #{issuer := Issuer}) ->
+    {fail, {bad_cert, root_cert_expired}};
 validate(_, {bad_cert, _} = Reason, _) ->
     {fail, Reason};
 validate(Cert, valid, UserState) ->
@@ -207,16 +271,21 @@ extensions_list(Extensions) ->
     Extensions.
 
 %%--------------------------------------------------------------------
--spec public_key_type(term()) -> rsa | dsa | ec.
+-spec public_key_type(term()) -> rsa | rsa_pss_pss | dsa | ecdsa | eddsa.
 %%
 %% Description:
 %%--------------------------------------------------------------------
+public_key_type(?'id-RSASSA-PSS') ->
+    rsa_pss_pss;
 public_key_type(?'rsaEncryption') ->
     rsa;
 public_key_type(?'id-dsa') ->
     dsa;
 public_key_type(?'id-ecPublicKey') ->
-    ec.
+    ecdsa;
+public_key_type(Oid) ->
+    {_, Sign} = public_key:pkix_sign_types(Oid),
+    Sign.
 
 %%--------------------------------------------------------------------
 -spec foldl_db(fun(), db_handle() | {extracted, list()}, list()) ->
@@ -229,10 +298,96 @@ foldl_db(IsIssuerFun, CertDbHandle, []) ->
 foldl_db(IsIssuerFun, _, [_|_] = ListDb) ->
     lists:foldl(IsIssuerFun, issuer_not_found, ListDb).
 
+find_cross_sign_root_paths([], _CertDbHandle, _CertDbRef, _InvalidatedList) ->
+    [];
+find_cross_sign_root_paths([_ | Rest] = Path, CertDbHandle, CertDbRef, InvalidatedList) ->
+    case find_alternative_root(Path, CertDbHandle, CertDbRef, InvalidatedList) of
+        unknown_ca ->
+            find_cross_sign_root_paths(Rest, CertDbHandle, CertDbRef, InvalidatedList);
+        Root ->
+            [{Root, Path}]
+    end.
+
+handle_cert_auths(Chain, [], _, _) ->
+    %% If we have no authorities extension (or corresponding
+    %% 'certificate_authorities' in the certificate request message in
+    %% TLS-1.2 is empty) to check we just accept first choice.
+    {ok, Chain};
+handle_cert_auths([Cert], CertAuths, CertDbHandle, CertDbRef) ->
+    case certificate_chain(Cert, CertDbHandle, CertDbRef, [], both) of
+        {ok, {_, [Cert | _] = EChain}, {_, [_ | DCerts]}}  ->
+            case cert_auth_member(cert_issuers(DCerts), CertAuths) of
+                true ->
+                    {ok, EChain};
+                false ->
+                    {error, EChain, not_in_auth_domain}
+            end;
+        _ ->
+            {ok, [Cert]}
+    end;
+handle_cert_auths([_ | Certs] = EChain, CertAuths, _, _) ->
+    case cert_auth_member(cert_issuers(Certs), CertAuths) of
+        true ->
+            {ok, EChain};
+        false ->
+            {error, EChain, not_in_auth_domain}
+    end.
+
+available_cert_key_pairs(CertKeyGroups) ->
+    %% To be able to find possible TLS session pre TLS-1.3
+    %% that may be reused. At this point the version is
+    %% not negotiated.
+    RevAlgos = [dsa, rsa, rsa_pss_pss, ecdsa],
+    cert_key_group_to_list(RevAlgos, CertKeyGroups, []).
+
+%% Create the prioritized list of cert key pairs that
+%% are availble for use in the negotiated version
+available_cert_key_pairs(CertKeyGroups, {3, 4}) ->
+    RevAlgos = [rsa, rsa_pss_pss, ecdsa, eddsa],
+    cert_key_group_to_list(RevAlgos, CertKeyGroups, []);
+available_cert_key_pairs(CertKeyGroups, {3, 3}) ->
+     RevAlgos = [dsa, rsa, rsa_pss_pss, ecdsa],
+    cert_key_group_to_list(RevAlgos, CertKeyGroups, []);
+available_cert_key_pairs(CertKeyGroups, {3, N}) when N < 3->
+    RevAlgos = [dsa, rsa, ecdsa],
+    cert_key_group_to_list(RevAlgos, CertKeyGroups, []).
+
+cert_key_group_to_list([], _, Acc) ->
+    final_group_list(Acc);
+cert_key_group_to_list([Algo| Rest], CertKeyGroups, Acc) ->
+    CertKeyPairs = maps:get(Algo, CertKeyGroups, []),
+    cert_key_group_to_list(Rest, CertKeyGroups, CertKeyPairs ++ Acc).
+
+final_group_list([]) ->
+    [#{certs => [[]], private_key => #{}}];
+final_group_list(List) ->
+    List.
+
 %%--------------------------------------------------------------------
 %%% Internal functions
 %%--------------------------------------------------------------------
-certificate_chain(OtpCert, BinCert, CertDbHandle, CertsDbRef, Chain, ListDb) ->
+encoded_chain(#cert{der=Cert}, Certs) ->
+    {Cert, [C || #cert{der=C} <- Certs]};
+encoded_chain(Res, Certs) ->
+    {Res, [OtpC || #cert{der=OtpC} <- Certs]}.
+
+decoded_chain(#cert{otp=OtpCert}, Certs) ->
+    {OtpCert, [OtpC || #cert{otp=OtpC} <- Certs]};
+decoded_chain(Res, Certs) ->
+    {Res, [OtpC || #cert{otp=OtpC} <- Certs]}.
+
+chain_result(Root0, Chain0, encoded) ->
+    {Root, Chain} = encoded_chain(Root0, Chain0),
+    {ok, Root, Chain};
+chain_result(Root0, Chain0, decoded) ->
+    {Root, Chain} = decoded_chain(Root0, Chain0),
+    {ok, Root, Chain};
+chain_result(Root0, Chain0, both) ->
+    {ERoot, EChain} = encoded_chain(Root0, Chain0),
+    {DRoot, DChain} = decoded_chain(Root0, Chain0),
+    {ok, {ERoot, EChain}, {DRoot, DChain}}.
+
+build_certificate_chain(#cert{otp=OtpCert}=Cert, CertDbHandle, CertsDbRef, Chain, ListDb) ->
     IssuerAndSelfSigned = 
 	case public_key:pkix_is_self_signed(OtpCert) of
 	    true ->
@@ -240,16 +395,15 @@ certificate_chain(OtpCert, BinCert, CertDbHandle, CertsDbRef, Chain, ListDb) ->
 	    false  ->
 		{public_key:pkix_issuer_id(OtpCert, other), false}
 	end,
-    
-    case IssuerAndSelfSigned of 
+    case IssuerAndSelfSigned of
 	{_, true = SelfSigned} ->
 	    do_certificate_chain(CertDbHandle, CertsDbRef, Chain, ignore, ignore, SelfSigned, ListDb);
 	{{error, issuer_not_found}, SelfSigned} ->
-	    case find_issuer(OtpCert, BinCert, CertDbHandle, CertsDbRef, ListDb) of
+	    case find_issuer(Cert, CertDbHandle, CertsDbRef, ListDb, []) of
 		{ok, {SerialNr, Issuer}} ->
 		    do_certificate_chain(CertDbHandle, CertsDbRef, Chain,
 					 SerialNr, Issuer, SelfSigned, ListDb);
-		_ ->
+		_Err ->
 		    %% Guess the the issuer must be the root
 		    %% certificate. The verification of the
 		    %% cert chain will fail if guess is
@@ -259,31 +413,48 @@ certificate_chain(OtpCert, BinCert, CertDbHandle, CertsDbRef, Chain, ListDb) ->
 	{{ok, {SerialNr, Issuer}}, SelfSigned} -> 
 	    do_certificate_chain(CertDbHandle, CertsDbRef, Chain, SerialNr, Issuer, SelfSigned, ListDb)
     end.
-  
-do_certificate_chain(_, _, [RootCert | _] = Chain, _, _, true, _) ->	  
-    {ok, RootCert, lists:reverse(Chain)};		      
+
+do_certificate_chain(_, _, [RootCert | _] = Chain, _, _, true, _) ->
+    {ok, RootCert, lists:reverse(Chain)};
 
 do_certificate_chain(CertDbHandle, CertsDbRef, Chain, SerialNr, Issuer, _, ListDb) ->
     case ssl_manager:lookup_trusted_cert(CertDbHandle, CertsDbRef,
-						SerialNr, Issuer) of
-	{ok, {IssuerCert, ErlCert}} ->
-	    ErlCert = public_key:pkix_decode_cert(IssuerCert, otp),
-	    certificate_chain(ErlCert, IssuerCert, 
-			      CertDbHandle, CertsDbRef, [IssuerCert | Chain], ListDb);
+                                         SerialNr, Issuer) of
+	{ok, Cert} ->
+	    build_certificate_chain(Cert, CertDbHandle, CertsDbRef, [Cert | Chain], ListDb);
 	_ ->
 	    %% The trusted cert may be obmitted from the chain as the
 	    %% counter part needs to have it anyway to be able to
 	    %% verify it.
-	    {ok, undefined, lists:reverse(Chain)}		      
+	    {ok, undefined, lists:reverse(Chain)}
     end.
 
+find_alternative_root([Cert | _], CertDbHandle, CertDbRef, InvalidatedList) ->
+    case find_issuer(Cert, CertDbHandle, CertDbRef, [], InvalidatedList) of
+        {error, issuer_not_found} ->
+            unknown_ca;
+        {ok, {SerialNr, IssuerId}} ->
+            case ssl_manager:lookup_trusted_cert(CertDbHandle, CertDbRef, SerialNr, IssuerId) of
+                undefined ->
+                    unknown_ca;
+                {ok, #cert{otp = OtpIssuer}} ->
+                    case public_key:pkix_is_self_signed(OtpIssuer) of
+                        true ->
+                            OtpIssuer;
+                        false ->
+                            unknown_ca
+                    end
+            end
+    end.
 
-find_issuer(OtpCert, BinCert, CertDbHandle, CertsDbRef, ListDb) ->
+find_issuer(#cert{der=DerCert, otp=OtpCert}, CertDbHandle, CertsDbRef, ListDb, InvalidatedList) ->
     IsIssuerFun =
-	fun({_Key, {_Der, #'OTPCertificate'{} = ErlCertCandidate}}, Acc) ->
+	fun({_Key, #cert{otp=ErlCertCandidate}}, Acc) ->
 		case public_key:pkix_is_issuer(OtpCert, ErlCertCandidate) of
 		    true ->
-			case verify_cert_signer(BinCert, ErlCertCandidate#'OTPCertificate'.tbsCertificate) of
+			case verify_cert_signer(DerCert, ErlCertCandidate#'OTPCertificate'.tbsCertificate)
+                            andalso (not lists:member(ErlCertCandidate, InvalidatedList))
+                        of
 			    true ->
 				throw(public_key:pkix_issuer_id(ErlCertCandidate, self));
 			    false ->
@@ -334,6 +505,12 @@ public_key(#'OTPSubjectPublicKeyInfo'{algorithm = #'PublicKeyAlgorithm'{algorith
 									parameters = Params},
 				      subjectPublicKey = Point}) ->
     {Point, Params};
+public_key(#'OTPSubjectPublicKeyInfo'{algorithm = #'PublicKeyAlgorithm'{algorithm = ?'id-Ed25519'},
+				      subjectPublicKey = Point}) ->
+    {Point, {namedCurve, ?'id-Ed25519'}};
+public_key(#'OTPSubjectPublicKeyInfo'{algorithm = #'PublicKeyAlgorithm'{algorithm = ?'id-Ed448'},
+				      subjectPublicKey = Point}) ->
+    {Point, {namedCurve, ?'id-Ed448'}};
 public_key(#'OTPSubjectPublicKeyInfo'{algorithm = #'PublicKeyAlgorithm'{algorithm = ?'rsaEncryption'}, 
 				      subjectPublicKey = Key}) ->
     Key;
@@ -346,12 +523,12 @@ public_key(#'OTPSubjectPublicKeyInfo'{algorithm = #'PublicKeyAlgorithm'{algorith
 				      subjectPublicKey = Key}) ->
     {Key, Params}.
 
-other_issuer(OtpCert, BinCert, CertDbHandle, CertDbRef) ->
+other_issuer(#cert{otp=OtpCert}=Cert, CertDbHandle, CertDbRef) ->
     case public_key:pkix_issuer_id(OtpCert, other) of
 	{ok, IssuerId} ->
 	    {other, IssuerId};
 	{error, issuer_not_found} ->
-	    case find_issuer(OtpCert, BinCert, CertDbHandle, CertDbRef, []) of
+	    case find_issuer(Cert, CertDbHandle, CertDbRef, [], []) of
 		{ok, IssuerId} ->
 		    {other, IssuerId};
 		Other ->
@@ -397,7 +574,8 @@ verify_cert_extensions(Cert, UserState, [], _) ->
     {valid, UserState#{issuer => Cert}};
 verify_cert_extensions(Cert, #{ocsp_responder_certs := ResponderCerts,
                                ocsp_state := OscpState,
-                               issuer := Issuer} = UserState, [#certificate_status{response = OcspResponsDer} | Exts], Context) ->
+                               issuer := Issuer} = UserState, 
+                       [#certificate_status{response = OcspResponsDer} | Exts], Context) ->
     #{ocsp_nonce := Nonce} = OscpState,
     case public_key:pkix_ocsp_validate(Cert, Issuer, OcspResponsDer, ResponderCerts, Nonce) of
         valid ->
@@ -406,28 +584,42 @@ verify_cert_extensions(Cert, #{ocsp_responder_certs := ResponderCerts,
             {fail, Status}
     end;
 verify_cert_extensions(Cert, UserState, [_|Exts], Context) ->
-    %% Skip unknow extensions!
+    %% Skip unknown extensions!
     verify_cert_extensions(Cert, UserState, Exts, Context).
 
 verify_sign(_, #{version := {_, Minor}}) when Minor < 3 ->
     %% This verification is not applicable pre TLS-1.2 
     true; 
-verify_sign(Cert, #{signature_algs := SignAlgs,
+verify_sign(Cert, #{version := {3, 3},
+                    signature_algs := SignAlgs,
                     signature_algs_cert := undefined}) ->
-    is_supported_signature_algorithm(Cert, SignAlgs); 
-verify_sign(Cert, #{signature_algs_cert := SignAlgs}) ->
-    is_supported_signature_algorithm(Cert, SignAlgs).
+    is_supported_signature_algorithm_1_2(Cert, SignAlgs);
+verify_sign(Cert, #{version := {3, 3},
+                    signature_algs_cert := SignAlgs}) ->
+    is_supported_signature_algorithm_1_2(Cert, SignAlgs);
+verify_sign(Cert, #{version := {3, 4},
+                    signature_algs := SignAlgs,
+                    signature_algs_cert := undefined}) ->
+    is_supported_signature_algorithm_1_3(Cert, SignAlgs);
+verify_sign(Cert, #{version := {3, 4},
+                    signature_algs_cert := SignAlgs}) ->
+    is_supported_signature_algorithm_1_3(Cert, SignAlgs).
 
-is_supported_signature_algorithm(#'OTPCertificate'{signatureAlgorithm = 
-                                                       #'SignatureAlgorithm'{algorithm = ?'id-dsa-with-sha1'}}, [{_,_}|_] = SignAlgs) ->   
+is_supported_signature_algorithm_1_2(#'OTPCertificate'{signatureAlgorithm =
+                                                           #'SignatureAlgorithm'{algorithm = ?'id-dsa-with-sha1'}},
+                                     SignAlgs) ->
     lists:member({sha, dsa}, SignAlgs);
-is_supported_signature_algorithm(#'OTPCertificate'{signatureAlgorithm = SignAlg}, [{_,_}|_] = SignAlgs) ->   
+is_supported_signature_algorithm_1_2(#'OTPCertificate'{signatureAlgorithm =
+                                                           #'SignatureAlgorithm'{algorithm = ?'id-RSASSA-PSS'}} = Cert,
+                                     SignAlgs) ->
+    is_supported_signature_algorithm_1_3(Cert, SignAlgs);
+is_supported_signature_algorithm_1_2(#'OTPCertificate'{signatureAlgorithm = SignAlg}, SignAlgs) ->
     Scheme = ssl_cipher:signature_algorithm_to_scheme(SignAlg),
     {Hash, Sign, _ } = ssl_cipher:scheme_to_components(Scheme),
-    lists:member({pre_1_3_hash(Hash), pre_1_3_sign(Sign)}, SignAlgs);
-is_supported_signature_algorithm(#'OTPCertificate'{signatureAlgorithm = SignAlg}, SignAlgs) ->   
+    ssl_cipher:is_supported_sign({pre_1_3_hash(Hash), pre_1_3_sign(Sign)}, ssl_cipher:signature_schemes_1_2(SignAlgs)).
+is_supported_signature_algorithm_1_3(#'OTPCertificate'{signatureAlgorithm = SignAlg}, SignAlgs) ->
     Scheme = ssl_cipher:signature_algorithm_to_scheme(SignAlg),
-    lists:member(Scheme, SignAlgs).
+    ssl_cipher:is_supported_sign(Scheme, SignAlgs).
 
 pre_1_3_sign(rsa_pkcs1) ->
     rsa;
@@ -443,10 +635,10 @@ paths(Chain, CertDbHandle) ->
 
 paths([Root], _, _, Path) ->
     [[Root | Path]];
-paths([Cert1, Cert2 | Rest], Chain, CertDbHandle, Path) ->
-    case public_key:pkix_is_issuer(Cert1, Cert2) of
+paths([#cert{otp=C1}=Cert1, #cert{otp=C2}=Cert2 | Rest], Chain, CertDbHandle, Path) ->
+    case public_key:pkix_is_issuer(C1, C2) of
         true ->
-            %% Chain orded so far
+            %% Chain ordered so far
             paths([Cert2 | Rest], Chain, CertDbHandle, [Cert1 | Path]);
         false ->
             %% Chain is unorded and/or contains extraneous certificates
@@ -460,40 +652,40 @@ unorded_or_extraneous([Peer | UnorderedChain], CertDbHandle) ->
               end,
               ChainCandidates).
 
-path_candidate(Peer, ChainCandidateCAs, CertDbHandle) ->
-    {ok,  ExtractedCerts} = ssl_pkix_db:extract_trusted_certs({der, ChainCandidateCAs}),
+path_candidate(Cert, ChainCandidateCAs, CertDbHandle) ->
+    {ok,  ExtractedCerts} = ssl_pkix_db:extract_trusted_certs({der_otp, ChainCandidateCAs}),
     %% certificate_chain/4 will make sure the chain is ordered
-    case certificate_chain(Peer, CertDbHandle, ExtractedCerts, []) of
+    case build_certificate_chain(Cert, CertDbHandle, ExtractedCerts, [Cert], []) of
         {ok, undefined, Chain} ->
             lists:reverse(Chain);
         {ok, Root, Chain} ->
             [Root | lists:reverse(Chain)]
     end.
 
-handle_partial_chain([IssuerCert| Rest] = Path, PartialChainHandler, CertDbHandle, CertDbRef) ->
-    case public_key:pkix_is_self_signed(IssuerCert) of
+handle_partial_chain([#cert{der=DERIssuerCert, otp=OtpIssuerCert}=Cert| Rest] = Path, PartialChainHandler,
+                     CertDbHandle, CertDbRef) ->
+    case public_key:pkix_is_self_signed(OtpIssuerCert) of
         true -> %% IssuerCert = ROOT (That is ROOT was included in chain)
-            {ok, {SerialNr, IssuerId}} = public_key:pkix_issuer_id(IssuerCert, self),
+            {ok, {SerialNr, IssuerId}} = public_key:pkix_issuer_id(OtpIssuerCert, self),
             case ssl_manager:lookup_trusted_cert(CertDbHandle, CertDbRef, SerialNr, IssuerId) of
-                {ok, {IssuerCert, _}} -> %% Match sent ROOT to trusted ROOT 
-                    maybe_shorten_path(Path, PartialChainHandler, {IssuerCert, Rest});
+                {ok, #cert{der=DERIssuerCert}} -> %% Match sent ROOT to trusted ROOT
+                    maybe_shorten_path(Path, PartialChainHandler, {Cert, Rest});
                 {ok, _} -> %% Did not match trusted ROOT
                     maybe_shorten_path(Path, PartialChainHandler, {invalid_issuer, Path});
                 _ ->
                     maybe_shorten_path(Path, PartialChainHandler, {unknown_ca, Path})
             end;
         false ->
-            OTPCert = public_key:pkix_decode_cert(IssuerCert, otp),
-            case other_issuer(OTPCert, IssuerCert, CertDbHandle, CertDbRef) of
+            case other_issuer(Cert, CertDbHandle, CertDbRef) of
                 {other, {SerialNr, IssuerId}} ->
                     case ssl_manager:lookup_trusted_cert(CertDbHandle, CertDbRef, SerialNr, IssuerId) of
-                        {ok, {NewIssuerCert, _}} ->  
-                            case public_key:pkix_is_self_signed(NewIssuerCert) of
+                        {ok, #cert{otp=NewOtp}=NewCert} ->
+                            case public_key:pkix_is_self_signed(NewOtp) of
                                 true -> %% NewIssuerCert is a trusted ROOT cert
-                                    maybe_shorten_path([NewIssuerCert | Path], PartialChainHandler, {NewIssuerCert, Path});
+                                    maybe_shorten_path([NewCert | Path], PartialChainHandler, {NewCert, Path});
                                 false ->
-                                    maybe_shorten_path([NewIssuerCert | Path], PartialChainHandler, 
-                                                       {unknown_ca, [NewIssuerCert | Path]})
+                                    maybe_shorten_path([NewCert | Path], PartialChainHandler, 
+                                                       {unknown_ca, [NewCert | Path]})
                             end;
                         _ ->
                             maybe_shorten_path(Path, PartialChainHandler, {unknown_ca, Path})
@@ -508,33 +700,36 @@ maybe_shorten_path(Path, PartialChainHandler, Default) ->
     %% certificate path to be validated with
     %% public_key:pkix_path_validation by letting
     %% the user put its trust in an intermidate cert
-    %% from the certifcate chain sent by the peer.
-    try PartialChainHandler(Path) of
+    %% from the certificate chain sent by the peer.
+    DerCerts = [Der || #cert{der=Der} <- Path],
+    try PartialChainHandler(DerCerts) of
         {trusted_ca, Root} ->
-            new_trusteded_path(Root, Path, Default);
+            new_trusted_path(Root, Path, Default);
         unknown_ca ->
             Default
     catch _:_ ->
             Default
     end.
 
-new_trusteded_path(DerCert, [DerCert | Chain], _) ->
-    {DerCert, Chain};
-new_trusteded_path(DerCert, [_ | Rest], Default) ->
-    new_trusteded_path(DerCert, Rest, Default);
-new_trusteded_path(_, [], Default) ->
+new_trusted_path(DerCert, [#cert{der=DerCert}=Cert | Path], _) ->
+    {Cert, Path};
+new_trusted_path(DerCert, [_ | Rest], Default) ->
+    new_trusted_path(DerCert, Rest, Default);
+new_trusted_path(_, [], Default) ->
     %% User did not pick a cert present 
     %% in the cert chain so ignore
     Default.
 
-handle_incomplete_chain([PeerCert| _] = Chain0, PartialChainHandler, Default, CertDbHandle, CertDbRef) ->
-    %% We received an incomplete chain, that is not all certs expected to be present are present. 
-    %% See if we have the certificates to rebuild it. 
-    case certificate_chain(PeerCert, CertDbHandle, CertDbRef) of
-        {ok, _, [PeerCert | _] = Chain} when Chain =/= Chain0 -> %% Chain candidate found          
-            case lists:prefix(Chain0, Chain) of 
+handle_incomplete_chain([#cert{}=Peer| _] = Chain0, PartialChainHandler, Default, CertDbHandle, CertDbRef) ->
+    %% We received an incomplete chain, that is not all certs expected to be present are present.
+    %% See if we have the certificates to rebuild it.
+    case build_certificate_chain(Peer, CertDbHandle, CertDbRef, [Peer], []) of
+        {ok, _, [Peer | _] = ChainCandidate} when ChainCandidate =/= Chain0 -> %% Chain candidate found
+            case lists:prefix(Chain0, ChainCandidate) of
                 true ->
-                    handle_partial_chain(lists:reverse(Chain), PartialChainHandler, CertDbHandle, CertDbRef);
+                    {Root, Chain} = handle_partial_chain(lists:reverse(ChainCandidate), PartialChainHandler, 
+                                                         CertDbHandle, CertDbRef),
+                    {Root, Chain};
                 false ->
                     Default
             end;
@@ -547,7 +742,7 @@ extraneous_chains(Certs) ->
     %% subject field we should create a list of possible chain certs
     %% for each such cert. Only one chain, if any, should be
     %% verifiable using available ROOT certs.
-    Subjects = [{subject(Cert), Cert} || Cert <- Certs],
+    Subjects = [{subject(OTP), Cert} || #cert{otp=OTP} = Cert <- Certs],
     Duplicates = find_duplicates(Subjects),
     %% Number of certs with duplicates (same subject) has been limited
     %% to 4 and the maximum number of combinations is limited to 16.
@@ -571,7 +766,7 @@ build_candidates([H|T], Map, Duplicates, Combinations, Max, Acc0) ->
 		_Else ->
 		    Acc = [[Cert|L] || Cert <- Certs, L <- Acc0],
 		    build_candidates(T, Map, Duplicates - 1, Combinations * Counter, Max, Acc)
-		end;
+            end;
 	{[Cert|_Throw], _Counter} ->
 	    case Acc0 of
 		[] ->
@@ -600,3 +795,29 @@ subject(Cert) ->
     {_Serial,Subject} = public_key:pkix_subject_id(Cert),
     Subject.
 
+issuer(Cert) ->
+    case public_key:pkix_is_self_signed(Cert) of
+        true ->
+            subject(Cert);
+        false ->
+            case is_binary(Cert) of
+                true ->
+                    #'OTPCertificate'{tbsCertificate = TBSCert} = public_key:pkix_decode_cert(Cert, otp),
+                    public_key:pkix_normalize_name(TBSCert#'OTPTBSCertificate'.issuer);
+                false ->
+                    #'OTPCertificate'{tbsCertificate = TBSCert} = Cert,
+                    public_key:pkix_normalize_name(TBSCert#'OTPTBSCertificate'.issuer)
+            end
+    end.
+
+cert_issuers([], Acc) ->
+    Acc;
+cert_issuers([Cert | Rest], Acc) ->
+    cert_issuers(Rest, [issuer(Cert) | Acc]).
+
+cert_issuers(OTPCerts) ->
+    cert_issuers(OTPCerts, []).
+
+cert_auth_member(ChainSubjects, CertAuths) ->
+    CommonAuthorities = sets:intersection(sets:from_list(ChainSubjects), sets:from_list(CertAuths)),
+    not sets:is_empty(CommonAuthorities).
